from ..core import Task


class match(Task):
    """
    Filters only matching files.

    NOTE: Matching is done from the right.
    """

    def __init__(self, pattern):
        self.pattern = pattern

    def process(self, inputs):
        for t, c, m in inputs:
            if t.match(self.pattern):
                yield t, c, m


class exclude(Task):
    """
    Excludes matching files.

    NOTE: Matching is done from the right.
    """

    def __init__(self, pattern):
        self.pattern = pattern

    def process(self, inputs):
        for t, c, m in inputs:
            if not t.match(self.pattern):
                yield t, c, m



from ..core import ExternalCommandTask


class external_command(ExternalCommandTask):
    def process(self, inputs):
        for t, c, m in inputs:
            yield t, [self.run_external(self.cmd, input=b''.join(c))], m



import subprocess
from itertools import tee

from click import echo, secho, style
from more_itertools import consume

from .exceptions import ExternalCommandError, IronTaskError
from .runtime import verbose
from .utils import chdir


def dictmerge(*dicts, **kwargs):
    """
    Merges an arbitrary number of dicts and kwargs into a single (new) dict.
    Inputs are unaffected.

    Usage:

        >>> d1 = {'a': 1}
        >>> d2 = {'b': 2}
        >>> m = dictmerge(d1, d2, moar=3)
        >>> sorted(m.items())
        [('a', 1), ('b', 2), ('moar', 3)]

    """
    result = {}
    for d in dicts:
        result.update(d)
    result.update(kwargs)
    return result


class Task:
    def __rshift__(self, other):
        return self.pipe(other)

    def __and__(self, other):
        return CombineOutputTask(self, other)

    def log(self, message):
        if verbose == True:  # noqa
            self.info(message)

    def info(self, message):
        self._info('>>> [{}] {}'.format(self.name, message))

    def _info(self, message):
        print(message)

    @property
    def name(self):
        return self.__class__.__name__

    def pipe(self, other):
        return ChainedTask(self, other)

    def __iter__(self):
        return self.process([])

    def format_for_exception(self, cause):
        s = str(self)
        if self == cause:
            return style(s, fg='red')
        return s

    def __str__(self):
        return self.name

    def run(self):
        try:
            # Iterating over the task triggers all the behaviour
            consume(self)
        except IronTaskError as e:
            echo('Exception occurred while processing:')
            echo(self.format_for_exception(cause=e.task))
            secho(str(e), fg='red')


class ChainedTask(Task):
    def __init__(self, task1, task2):
        self.task1 = task1
        self.task2 = task2

    def process(self, inputs):
        return self.task2.process(self.task1.process(inputs))

    def __str__(self):
        return '{} >> {}'.format(self.task1, self.task2)

    def format_for_exception(self, cause):
        return ('{} >> {}'.format(self.task1.format_for_exception(cause),
                                  self.task2.format_for_exception(cause)))


class CombineOutputTask(Task):
    def __init__(self, task1, task2, **kwargs):
        super().__init__(**kwargs)
        self.task1 = task1
        self.task2 = task2

    def process(self, inputs):
        inputs1, inputs2 = [], []
        for t, c, m in inputs:
            self.log('t = {}'.format(t))
            self.log('c = {}'.format(c))
            c1, c2 = tee(c)
            inputs1.append((t, c1, dictmerge(m)))
            inputs2.append((t, c2, dictmerge(m)))

        self.log('inputs1 = {}'.format(inputs1))
        self.log('inputs2 = {}'.format(inputs2))
        yield from self.task1.process(inputs1)
        yield from self.task2.process(inputs2)

    def __str__(self):
        return '({} & {})'.format(self.task1, self.task2)

    def format_for_exception(self, cause):
        return ('({} & {})'.format(self.task1.format_for_exception(cause),
                                   self.task2.format_for_exception(cause)))


class ExternalCommandTask(Task):
    def run_external(self, cmd, *, cwd=None, input=b''):
        with chdir(cwd):
            try:
                return subprocess.check_output(cmd, input=input)
            except subprocess.CalledProcessError as e:
                raise ExternalCommandError('Error while running {}'.format(cmd),
                                           cmd=cmd,
                                           details=str(e),
                                           task=self, cwd=cwd) from e



import humanize
import os
import pprint
from itertools import tee

from click import echo, secho, style
from pathlib import Path

from ..core import Task
from ..exceptions import IronError
from ..io import lazy_read, write_to_file
from ..runtime import verbose


class dump(Task):
    """
    Dumps (echoes) the contents of all streams.

    Synchronizes and buffers all stream contents before proceding, unless
    sync=False is used.
    """
    def __init__(self, sync=True):
        self.sync = sync

    def process(self, inputs):
        if self.sync:
            inputs = list(inputs)

        for t, c, m in inputs:
            c1, c2 = tee(c)
            secho(str(t), fg='magenta')
            secho(pprint.pformat(m, depth=1), fg='yellow')
            if verbose == True:  # TODO: ugh, use g.verbose  # noqa
                echo(b''.join(c1))
            yield t, c2, m


class glob(Task):
    """
    Globs from the current directory using the given pattern(s).

    Multiple patterns may be given.  If some files match multiple patterns,
    they will only be yielded once.  This property may be put to use
    conveniently for example when the order of concatenation is important for
    *some* files.  Example::

        glob('src/js/main.js',  # first output main.js
             'src/js/**/*.js'   # then everything else (main.js will not be included again)
        )

    This will even work when the pattern is specified a little different::

        # even with a different prefix, glob knows it's the same file
        glob('src/js/main.js',
             './src/js/**/*.js'
        )

    """

    def __init__(self, *patterns):
        self.patterns = patterns

    def process(self, inputs):
        root = Path('.')

        # Keep a set to filter out duplicates (matching multiple patterns)
        seen = set()
        for pattern in self.patterns:
            for t in root.glob(pattern):
                if t not in seen and t.is_file():
                    yield t, lazy_read(t), {}
                    seen.add(t)
        if not seen:
            raise IronError('{} did not output any files'.format(self))

    def __str__(self):
        return 'glob({})'.format(', '.join(map(repr, self.patterns)))


class log(Task):
    def __init__(self, message, sync=True):
        self.message = message
        self.sync = sync

    def process(self, inputs):
        if self.sync:
            inputs = list(inputs)
        self._info(self.message)
        yield from inputs


class noop(Task):
    def process(self, inputs):
        yield from inputs


class sync(Task):
    def process(self, inputs):
        inputs = list(inputs)
        yield from inputs


class walk(Task):
    """
    Given a root directory, yields each file underneath it, recursively
    iterating over all directories.

    Example::

        >>> t = walk('src') >> dump()
        >>> t.run()
        foo.js
        bar/baz.js
        bar/qux.js
        ...

        >>> t = walk('src', relative=False) >> dump()
        >>> t.run()
        src/foo.js
        src/bar/baz.js
        src/bar/qux.js
        ...

    """
    IGNORE = {'.DS_Store'}

    def __init__(self, *roots, relative=True):
        roots = list(map(Path, roots))
        for root in roots:
            assert root.is_dir(), 'Directory not found: {}'.format(root)
        self.roots = roots
        self.relative = relative

    def _iter_recursive(self, root):
        for path in root.iterdir():
            if path.is_file() or path.is_symlink():
                if path.name not in self.IGNORE:
                    yield path
            elif path.is_dir():
                yield from self._iter_recursive(path)

    def process(self, inputs):
        for root in self.roots:
            for path in self._iter_recursive(root):
                if self.relative:
                    yield path.relative_to(root), lazy_read(path), {}
                else:
                    yield path, lazy_read(path), {}


class write(Task):
    """
    Writes inputs to the file system.

    .. note::
       Since this is typically the last task in a chain, the default
       invocation of ``write()`` does not make a copy of its inputs and will
       simply emit empty contents for each file.

       If you want to produce intermediate writes or chain it to another task,
       use ``end=False``.  This will write the files to disk and pass through
       the content streams to the next process.
    """
    CLEARED_MSG = b'Output cleared by write(). Use write(end=False) to prevent this.'

    def __init__(self, root, silent=False, end=True):
        self.silent = silent
        self.end = end
        self.root = Path(root)

    def process(self, inputs):
        dirs = set()
        for t, c, m in inputs:
            concrete_path = self.root.joinpath(t)

            dir = concrete_path.parent
            if dir not in dirs:
                dirs.add(dir)
                if not dir.exists():
                    if not self.silent:
                        self.log('mkdir {}'.format(dir))
                    dir.mkdir(parents=True)

            if self.end:
                c1, c2 = c, [self.CLEARED_MSG]
            else:
                c1, c2 = tee(c)
            write_to_file(concrete_path, c1)
            stat_info = os.stat(str(concrete_path))
            num_bytes = stat_info.st_size
            num_bytes = humanize.naturalsize(num_bytes)
            if not self.silent:
                echo(style(str(concrete_path), fg='green') + ' written (' + style(num_bytes, fg='magenta') + ')')
            else:
                self.log(str(concrete_path) + ' written (' + num_bytes + ')')
            yield t, c2, m



import json
import subprocess

from pathlib import Path

from ..core import Task


class ng_annotate(Task):
    """Annotates static dependency injections for Angular source files."""

    def process(self, inputs):
        for t, c, m in inputs:
            cmd = ['ng-annotate', '-a', '-']
            output = subprocess.check_output(cmd, input=b''.join(c))
            yield t, [output], m


class ng_min(Task):
    """
    Rewrites Angular source files to be safe for minification.

    .. warning::
       ``ngmin`` does the same thing as ``ng-annotate`` tool, but the latter
       is much faster.  You can simply use ``ng_annotate()`` wherever you were
       using ``ng_min()``.
    """

    def process(self, inputs):
        for t, c, m in inputs:
            cmd = ['ngmin']
            output = subprocess.check_output(cmd, input=b''.join(c))
            yield t, [output], m


class ng_template_cache(Task):
    def __init__(self, target, module, create_module=False):
        self.module = module
        self.target = Path(target)
        self.create_module = create_module

    def lazy_template_cache(self, inputs):
        yield b'angular.module('
        yield json.dumps(self.module).encode('utf-8')
        if self.create_module:
            yield b', []'
        yield b').run(["$templateCache", function ($templateCache) {\n\n'

        for t, c, _ in inputs:
            yield b'    $templateCache.put('
            yield json.dumps(str(t)).encode('utf-8')
            yield b',\n'
            yield b'          '
            yield json.dumps(b''.join(c).decode('utf-8')).encode('utf-8')
            yield b');\n\n'

        yield b"\n}]);\n"

    def process(self, inputs):
        yield self.target, self.lazy_template_cache(inputs), {}



import hashlib
import json
from itertools import tee

from ..core import Task


class cachebust(Task):
    """
    Hashes all input files, and writes their hash value into the filename.

    Also outputs a manifest.json file, containing all the mappings.
    """
    def __init__(self, algorithm='sha512', length=7):
        self.algorithm = algorithm
        self.length = length

    def process(self, inputs):
        mapping = {}
        for t, c, m in inputs:
            c1, c2 = tee(c)
            h = getattr(hashlib, self.algorithm)()
            for chunk in c1:
                h.update(chunk)
            t2 = t.stem + '-' + h.hexdigest()[:self.length] + t.suffix
            mapping[str(t)] = str(t2)
            yield t2, c2, m
        yield 'manifest.json', [json.dumps(mapping).encode('utf-8')], {}



from ..core import Task


class minify_html(Task):
    """Minifies HTML input files."""

    def __init__(self, *htmlmin_args, **htmlmin_kwargs):
        try:
            from htmlmin import Minifier
        except ImportError as e:
            raise ImportError('To use minify_html(), run: pip install htmlmin') from e

        self.minifier = Minifier(*htmlmin_args, **htmlmin_kwargs)

    def lazy_minify(self, contents):
        yield self.minifier.minify(b''.join(contents).decode('utf-8')).encode('utf-8')

    def process(self, inputs):
        for t, c, m in inputs:
            yield t, self.lazy_minify(c), m



from gzip import GzipFile
from io import BytesIO

from ..core import Task


class gzip(Task):
    def lazy_gzip(self, input_stream):
        gzip_buffer = BytesIO()
        gzip_file = GzipFile(mode='wb', fileobj=gzip_buffer)
        try:
            for chunk in input_stream:
                gzip_file.write(chunk)
            gzip_file.flush()
        finally:
            gzip_file.close()
        yield gzip_buffer.getvalue()

    def process(self, inputs):
        for t, c, m in inputs:
            t2 = t.with_name(t.name + '.gz')
            yield t2, self.lazy_gzip(c), m



# flake8: noqa

# Generic tasks
from .basics import dump, glob, log, noop, sync, walk, write
from .io import external_command
from .filters import exclude, match
from .paths import directory, relative_to, rename_replace, strip_dirs
from .text import concat, contents_replace, footer, header, regex
from .templating import jinja2

# Specific tasks
from .angular import ng_annotate, ng_min, ng_template_cache
from .css import autoprefixer, cleancss, lessc
from .gzip import gzip
from .html import minify_html
from .javascript import coffee, traceur, uglify
from .markdown import markdown
from .png import optipng, pngquant
from .textile import textile
from .web import cachebust



from io import BytesIO

from ..core import ExternalCommandTask
from ..utils import write_to_tmp_dir
from .basics import walk


class optipng(ExternalCommandTask):
    """Optimize PNG files with optipng (lossless compression)."""

    def __init__(self, *, level=None):
        self.level = level

    def process(self, inputs):
        tmp_dir, paths = write_to_tmp_dir(inputs)

        cmd = ['optipng']
        if self.level:
            cmd += ['-o{}'.format(self.level)]
        cmd += [str(f) for f in paths]

        self.run_external(cmd, cwd=tmp_dir)

        return walk(tmp_dir)


class pngquant(ExternalCommandTask):
    """Optimize PNG files with pngquant (lossy compression)."""

    def __init__(self, *, quality):
        self.quality = '{}'.format(quality)

    def process(self, inputs):
        for t, c, m in inputs:
            cmd = ['pngquant']
            if self.quality:
                cmd += ['--quality', self.quality]
            cmd += ['--', '-']
            c2 = self.run_external(cmd, input=b''.join(c))
            yield t, BytesIO(c2), m



import sys
import time

from click import echo

from .commands import run_commands


_watch_rules = set()


def watch_forever(env, verbose):
    """
    Starts the watcher.  This is the meat driving the ``iron watch`` shell
    command.
    """
    try:
        from watchdog.events import FileDeletedEvent, PatternMatchingEventHandler
        from watchdog.observers import Observer
    except ImportError:
        echo('The watcher requires watchdog to be installed.')
        sys.exit(3)

    class CommandTrigger(PatternMatchingEventHandler):
        def __init__(self, patterns, cmdname, env, verbose):
            super().__init__(patterns=patterns, ignore_patterns=['.DS_Store'], ignore_directories=True)
            self.cmdname = cmdname
            self.env = env
            self.verbose = verbose

        def on_any_event(self, event):
            if isinstance(event, FileDeletedEvent):
                # HACK: My vim is a bit messed up and deletes a file before
                # writing it, meaning this will typically trigger 2 changes
                # for every save, and thus 2 rebuilds.  The proper way to
                # handle this is to batch the events for, say 0.1 seconds, and
                # only when nothing happens within this time window, execute
                # it.  I.e. always take a 0.1 delay, but avoid the double
                # trigger.
                echo('*Ignoring file deletion* {}'.format(event.src_path))
                return

            echo('>>> Running {}'.format(self.cmdname))
            run_commands([self.cmdname], self.env, self.verbose)

    observer = Observer(timeout=1.7)

    for name, paths, patterns in sorted(_watch_rules):
        handler = CommandTrigger(patterns, name, env, verbose)

        for path in paths:
            patstr = ', '.join(patterns) if patterns else 'any file'
            echo('Will trigger {} when {} changes in {}'.format(name, patstr, path))
            observer.schedule(handler, path, recursive=True)

    observer.start()
    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        print('Stopping file system watches...')
        observer.stop()
    observer.join()



from click import style


class IronError(Exception):
    pass


class IronFileNotFound(IronError):
    pass


class IronTaskError(IronError):
    def __init__(self, message, task):
        self.message = message
        self.task = task

    def __str__(self):
        return style(self.message, fg='red') + ' occurred in ' + style(self.task.name, fg='blue')


class ExternalCommandError(IronTaskError):
    def __init__(self, message, task, cmd, cwd, details=None):
        super().__init__(message, task)
        self.cmd = cmd
        self.cwd = cwd
        self.details = details

    def __str__(self):
        return (style(self.message, fg='red') + ' (from task ' +
                style(self.task.name, fg='blue') + ')\n' +
                'cmd was: ' + style(' '.join(self.cmd), fg='blue') + '\n' +
                'cwd was: ' + style(str(self.cwd), fg='blue') +
                '\n\nDetails:\n' +
                self.details)



def as_text(c):
    if isinstance(c, str):
        return c
    else:
        return b''.join(c).decode('utf-8')


def as_byte_stream(s):
    if not isinstance(s, str):
        raise TypeError('Expected a string, got {}'.format(s))
    yield s.encode('utf-8')



from .registry import _command_registry
from .watchers import _watch_rules


class command:
    """
    Registers a function in the ironfile as a command.

    Adding this to a function makes the function executable by running it on
    the commandline with ``iron``::

        from iron import command

        @task('ohai')
        def greet():
            print('ohai, world!')

        @task()
        def kthxbai():
            print('kthxbai, cruel world')

    The command name will be derived from the function, unless an alternative
    name is specified.
    """
    def __init__(self, name=None):
        self.name = name

    def __call__(self, f):
        name = self.name or f.__name__
        f._iron_cmdname = name
        _command_registry[name] = f
        return f


class watch:
    """
    Registers a watch pattern.
    """
    def __init__(self, path, pattern=None):
        self.path = path
        self.pattern = pattern

    def __call__(self, f):
        path = self.path
        if isinstance(path, str):
            path = [path]
        pat = self.pattern
        if pat is None:
            pat = []
        elif isinstance(pat, str):
            pat = [pat]
        _watch_rules.add((f._iron_cmdname, tuple(path), tuple(pat)))
        return f



import sys
from importlib import import_module
from shutil import rmtree

import click
from click import echo, pass_context

from pathlib import Path

from .commands import run_commands
from .exceptions import IronError, IronFileNotFound
from .registry import _command_registry
from .utils import IRON_TMP_ROOT
from .watchers import watch_forever


def find_ironfile():
    """
    Attempt to locate an ironfile.

    It tries to find a file named "ironfile.py" by searching parent dirs to
    the current working directory.
    """
    IRONFILE = 'ironfile.py'

    def iter_dirs():
        path = Path('.').absolute()
        yield path
        yield from path.parents

    def iter_files():
        for path in iter_dirs():
            yield path.joinpath(IRONFILE)

    tried = []
    for path in iter_files():
        if path.is_file():
            return path
        tried.append(path)

    raise IronFileNotFound('Could not find an ironfile.py. Tried: {}'.format(tried))


class add_to_load_path:
    """
    Adds a directory to the load path while the context is active, and
    restores the load path to whatever it was before the block was run.
    """
    def __init__(self, dir):
        self.dir = dir

    def __enter__(self):
        self.path_copy = sys.path[:]
        sys.path.insert(0, str(self.dir))

    def __exit__(self, *args):
        sys.path = self.path_copy[:]


def load_ironfile():
    """Finds and loads an ironfile."""
    ironfile = find_ironfile()
    dir = ironfile.parent

    # Pull a trick where we temporarily put the dir containing the ironfile on
    # the load path, then load the module, and remove it again.
    with add_to_load_path(dir):
        ironfile = import_module('ironfile')

    return ironfile


def print_usage(ctx):
    echo(ctx.get_help() + '\n')
    echo('Available commands:')
    for command in sorted(_command_registry):
        echo('  {:14s}  {}'.format(command, (_command_registry[command].__doc__ or '').strip()))
    echo('')


@click.command()
@click.option('-e', '--env', help='Target environment.', default='dev', metavar='<name>')
@click.option('-v', '--verbose', is_flag=True, help='Be verbose.')
@click.argument('commands', nargs=-1)
@pass_context
def cli(ctx, env, verbose, commands):
    load_ironfile()

    if not commands:
        print_usage(ctx)
        sys.exit(2)

    # Treat the 'watch' command special
    start_watching = 'watch' in commands
    commands = list(cmdname for cmdname in commands if cmdname != 'watch')

    # Check if all commands are known
    for cmdname in commands:
        if cmdname not in _command_registry:
            click.secho('Unknown command: "{}"'.format(cmdname), fg='red')
            sys.exit(2)

    try:
        run_commands(commands, env, verbose)

        if start_watching:
            try:
                watch_forever(env, verbose)
            except KeyboardInterrupt:
                click.echo('Stopping watcher...')
    except IronError as e:
        print(e)
    else:
        # Clean up temporary directory, if available
        if Path(IRON_TMP_ROOT).is_dir():
            rmtree(IRON_TMP_ROOT, ignore_errors=True)


if __name__ == '__main__':
    cli()



from ..core import Task


class jinja2(Task):
    def __init__(self, **context):
        try:
            from jinja2 import Template
        except ImportError as e:
            raise ImportError('To use jinja2() install Jinja2 first: pip install jinja2') from e
        self.template_cls = Template
        self.context = context

    def lazy_replace(self, c):
        text = b''.join(c).decode('utf-8')
        template = self.template_cls(text)
        output_text = template.render(**self.context)
        yield output_text.encode('utf-8')

    def process(self, inputs):
        for t, c, m in inputs:
            t = t.with_name(t.name.replace('.jinja2', ''))  # strip off any Jinja2 extension
            yield t, self.lazy_replace(c), m



from pathlib import Path

from ..core import Task


class directory(Task):
    def __init__(self, dirname):
        self.dirname = Path(dirname)

    def process(self, inputs):
        for t, c, m in inputs:
            yield self.dirname / t, c, m


class modify_path(Task):
    """Generic purpose task for changing file paths."""
    def __init__(self, func):
        self.func = func

    def process(self, inputs):
        for t, c, m in inputs:
            yield Path(self.func(t)), c, m


def rename_replace(substring, replacement):
    func = lambda path: str(path).replace(substring, replacement)
    return modify_path(func)


class relative_to(Task):
    def __init__(self, base):
        self.base = base

    def process(self, inputs):
        for t, c, m in inputs:
            try:
                t2 = t.relative_to(self.base)
            except ValueError:
                continue  # drop file
            yield t2, c, m


class strip_dirs(Task):
    """
    Flattens file names, by stripping off directory components from the files
    and keeping only the base names.

    Modifies target filenames.
    Does not modify stream contents.

    Example::

        foo.js               =>  foo.js
        foo/bar/baz/quux.js  =>  quux.js

    .. warning::
       Be aware that stripping off directory components may result in
       duplicate file names, for example ``foo/qux.js`` and ``bar/qux.js``
       both rename their target to ``qux.js``, which is fine while still
       processing the stream, but will result in undefined behaviour when
       writing the stream output to disk.
    """
    def process(self, inputs):
        for t, c, m in inputs:
            t2 = Path(t.name)
            yield t2, c, m



from io import BytesIO

from pathlib import Path

from ..core import ExternalCommandTask
from ..io import lazy_read
from ..utils import write_to_tmp_dir
from .basics import walk


class autoprefixer(ExternalCommandTask):
    def __init__(self, browsers=None):
        self.browsers = browsers

    def lazy_prefix(self, contents):
        cmd = ['postcss', '--use', 'autoprefixer']
        if self.browsers:
            cmd += ['--autoprefixer.browsers', self.browsers]

        output = self.run_external(cmd, input=b''.join(contents))
        yield output

    def process(self, inputs):
        for t, c, m in inputs:
            yield t, self.lazy_prefix(c), m


class cleancss(ExternalCommandTask):
    """Minifies CSS input files."""

    def __init__(self, *, s0=None, s1=None, keep_line_breaks=None):
        self.s0 = s0
        self.s1 = s1
        self.keep_line_breaks = keep_line_breaks

    def lazy_clean(self, c):
        cmd = ['cleancss']
        if self.s0:
            cmd += ['--s0']
        if self.s1:
            cmd += ['--s1']
        if self.keep_line_breaks:
            cmd += ['--keep-line-breaks']
        yield self.run_external(cmd, input=b''.join(c))

    def process(self, inputs):
        for t, c, m in inputs:
            if t.suffix != '.css':
                yield t, c, m
                continue

            yield t, self.lazy_clean(c), m


class lessc(ExternalCommandTask):
    """
    Compile LESS files to CSS.

    There are two modes to invoke the LESS compiler.

    The default mode passes the input .less files to lessc's stdin.  This
    method has the advantage of being slightly quicker, as no temporary files
    have to be created.

    The second method writes the input files to a temporary directory before
    invoking the less compiler, meaning you can use @import statements in the
    .less files.
    """

    def __init__(self, *files, line_numbers=None, source_map=False):
        self.files = files
        self.line_numbers = line_numbers
        self.source_map = source_map

    def process_via_stdin(self, inputs):
        for t, c, m in inputs:
            if t.suffix != '.less':
                yield t, c, m
                continue

            t2 = t.with_suffix('.css')
            t2_smap = t2.with_name(t2.name + '.map')

            cmd = ['lessc']
            if self.source_map:
                # TODO: Change this to write to a safer temporary directory
                cmd += ['--source-map=/tmp/_sm.map',
                        '--source-map-url=' + str(t2_smap.name)]
            if self.line_numbers:
                cmd += ['--line-numbers=' + self.line_numbers]
            cmd += ['-']

            output = self.run_external(cmd, input=b''.join(c))
            yield t2, BytesIO(output), m
            if self.source_map:
                yield t2_smap, lazy_read('/tmp/_sm.map'), m

    def process_via_tmp_dir(self, inputs):
        tmp_dir, paths = write_to_tmp_dir(inputs)

        base_cmd = ['lessc']
        if self.line_numbers:
            base_cmd += ['--line-numbers=' + self.line_numbers]

        if self.source_map:
            base_cmd += ['--source-map-less-inline',
                         '--source-map-map-inline']

        for f in self.files:
            output_path = Path(f).with_suffix('.css')

            self.run_external(base_cmd + [str(f), str(output_path)],
                              cwd=tmp_dir)

        for t, c, m in walk(tmp_dir):
            if t not in paths:  # filter out any input files, returning only the new files
                yield t, c, m

    def process(self, inputs):
        if self.files:
            return self.process_via_tmp_dir(inputs)
        else:
            return self.process_via_stdin(inputs)



from functools import partial

from ..core import Task


class markdown(Task):
    """Converts Markdown to HTML."""

    def __init__(self):
        try:
            from markdown import markdown as markdown_fn
        except ImportError as e:
            raise ImportError('To use markdown(), run: pip install markdown') from e

        from markdown.extensions.codehilite import CodeHiliteExtension
        hilite = CodeHiliteExtension(guess_lang=False)

        self.convert = partial(markdown_fn, extensions=[
            # 'markdown.extensions.codehilite',
            hilite,
            'markdown.extensions.fenced_code',
            'markdown.extensions.attr_list',
            'markdown.extensions.sane_lists',
            # 'markdown.extensions.tables',
            'markdown.extensions.headerid',
            'markdown.extensions.def_list',
        ])

    def lazy_convert(self, contents):
        yield self.convert(b''.join(contents).decode('utf-8')).encode('utf-8')

    def process(self, inputs):
        for t, c, m in inputs:
            t2 = t.with_suffix('.html')
            yield t2, self.lazy_convert(c), m



import re
from itertools import chain

from more_itertools import flatten
from pathlib import Path

from ..core import Task
from ._helpers import as_byte_stream, as_text


class concat(Task):
    """
    Concatenates input stream contents to the given output file, in the order
    that they appear in the inputs.

    Modifies target filename.
    Modifies stream contents.
    Destroys metadata.

    Example::

        foo.js, bar.js, qux.js  =>  all.js

    """
    def __init__(self, dest):
        self.dest = Path(dest)

    def process(self, inputs):
        yield self.dest, flatten(chain(c for _, c, _ in inputs)), {}


class contents_replace(Task):
    def __init__(self, substring, replacement):
        self.substring = substring
        self.replacement = replacement

    def lazy_replace(self, c):
        contents = b''.join(c).decode('utf-8').replace(self.substring, self.replacement).encode('utf-8')
        yield contents

    def process(self, inputs):
        for t, c, m in inputs:
            yield t, self.lazy_replace(c), m


class footer(Task):
    def __init__(self, bites):

        if isinstance(bites, str):
            bites = bites.encode('utf-8')
        if not isinstance(bites, bytes):
            raise TypeError('expected bytes')
        self.bites = bites

    def process(self, inputs):
        for t, c, m in inputs:
            yield t, chain(c, [self.bites]), m


class header(Task):
    def __init__(self, bites):

        if isinstance(bites, str):
            bites = bites.encode('utf-8')
        if not isinstance(bites, bytes):
            raise TypeError('expected bytes')
        self.bites = bites

    def process(self, inputs):
        for t, c, m in inputs:
            yield t, chain([self.bites], c), m


class regex(Task):
    """
    Takes a `pattern` (can be a compiled regex), and a `replacement`, and will
    perform the replacement on the contents of each input.

    Note: if you pass in the pattern as a string, then `re.MULTILINE` is
    assumed, since the contents of the file is passed to `.sub()` as a single
    string, not line-by-line.  If you don't want this, override the `flags`
    argument.

    Note: the `flags` argument is ignored when you pass in a pre-compiled
    regex.
    """
    def __init__(self, pattern, replacement, *, flags=re.MULTILINE):
        if isinstance(pattern, str):
            self.regex = re.compile(pattern, flags)
        elif hasattr(pattern, 'sub'):  # quacks like a compiled regexp
            self.regex = pattern
        else:
            raise TypeError('Expected a compiled regexp or a pattern, but got {}'.format(pattern))
        self.replacement = replacement

    def process(self, inputs):
        for t, c, m in inputs:
            contents = as_text(c)
            c2 = self.regex.sub(self.replacement, contents)
            yield t, as_byte_stream(c2), m



from ..core import Task


class textile(Task):
    """Converts Textile to HTML."""

    def __init__(self):
        try:
            from textile import textile as textile_fn
        except ImportError as e:
            raise ImportError('To use textile(), run: pip install textile') from e

        self.convert = textile_fn

    def lazy_convert(self, contents):
        yield self.convert(b''.join(contents).decode('utf-8')).encode('utf-8')

    def process(self, inputs):
        for t, c, m in inputs:
            t2 = t.with_suffix('.html')
            yield t2, self.lazy_convert(c), m



"""
iron: the powerful but simple build tool using flows.
"""
from setuptools import setup, find_packages


setup(
    name='iron',
    version='0.2.0',
    url='https://github.com/nvie/iron',
    license='BSD',
    author='Vincent Driessen',
    author_email='vincent@3rdcloud.com',
    description=__doc__.strip(),
    long_description=__doc__,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=['click', 'more_itertools', 'humanize'],
    entry_points={
        'console_scripts': [
            'iron = iron.cli:cli',
        ],
    },
    extras_require={
        'all': ['jinja2', 'watchdog', 'htmlmin'],
        'html': ['htmlmin'],
        'watch': ['watch'],
        'jinja2': ['jinja2'],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ]
)



"""
Helper functions for implementing Tasks more easily.
"""
import os
from shutil import rmtree
from tempfile import mkdtemp

from pathlib import Path

IRON_TMP_ROOT = '.iron-tmp'


class chdir:
    """
    Context manager to temporarily change the current directory.

    Usage::

        print(os.getcwd())  # prints /home/nvie
        with chdir('/tmp'):
            print(os.getcwd())  # prints /tmp
        print(os.getcwd())  # prints /home/nvie

    """
    def __init__(self, target):
        self._target = str(target) if target is not None else None  # supports Paths, too
        self._prev = None

    def __enter__(self):
        if self._target:
            self._prev = os.getcwd()
            os.chdir(self._target)

    def __exit__(self, *args):
        if self._target:
            os.chdir(self._prev)


def write_to_tmp_dir(inputs):
    """
    Helper function to write inputs to a temporary directory on disk.

    The files will be written to a temporary directory, where they can be
    consumed by any external process.

    Returns:
       A tuple of the form ``(tmp_dir_path, list_of_filenames)``, example::

           (PosixPath('.iron-tmp/sTu9wX3k1'),
            ['foo.js', 'bar/baz.js', 'bar/qux.js'])

    .. note::
       The temporary directory will be automatically cleaned up when all tasks
       are finished running.
    """

    from .lib import write

    dir = Path(IRON_TMP_ROOT)
    if not dir.exists():
        dir.mkdir(parents=True)

    tmp_dir = Path(mkdtemp(dir=str(dir)))
    paths = []
    for t, _, _ in write(tmp_dir, silent=True).process(inputs):
        paths.append(t)
    return tmp_dir, paths


def rm_rf(*paths):
    """
    Recursively removes the given directories or files, if they exist.

    Like running ``rm -rf <dir> [...]``
    """
    for path in paths:
        path = Path(path)
        if path.is_dir():
            rmtree(str(path), ignore_errors=True)
        elif path.is_file():
            path.unlink()



from pathlib import Path


def lazy_read(p):
    p = Path(p)
    with p.open('rb') as f:
        while True:
            chunk = f.read(4096)
            if chunk == b'':
                break
            yield chunk


def write_to_file(dest, input_stream):
    with Path(dest).open('wb') as f:
        for chunk in input_stream:
            if not isinstance(chunk, bytes):
                raise TypeError('bytes expected in input stream, got {}:'.format(type(chunk)))
            f.write(chunk)



from .core import Task
from .registry import _command_registry
from .runtime import _env_stack, _verbose_stack


def run_commands(commands, env, verbose):
    _env_stack.push(env)
    _verbose_stack.push(verbose)
    try:
        for cmdname in commands:
            rv = _command_registry[cmdname]()
            if isinstance(rv, Task):
                rv.run()
    finally:
        assert env == _env_stack.pop(), 'oops?'
        assert verbose == _verbose_stack.pop(), 'oops?'



import subprocess

from ..core import ExternalCommandTask, Task
from ..exceptions import ExternalCommandError
from ..utils import chdir, write_to_tmp_dir
from .basics import walk


class ExternalCommand:
    def __init__(self, cmd):
        self.cmd = cmd
        self.require_empty_stdout = False
        self.require_empty_stderr = False
        self.require_success = False

    def run(self):
        stdin = b''
        with subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE) as process:
            try:
                stdout, stderr = process.communicate(stdin)
            except:
                process.kill()
                process.wait()
                raise

            retcode = process.poll()

            if self.require_empty_stderr:
                if stderr:
                    raise ExternalCommandError(stderr.decode('utf-8'))

            if self.require_empty_stdout:
                if stdout:
                    raise ExternalCommandError(stdout.decode('utf-8'))

            if self.require_success:
                if retcode:
                    raise ExternalCommandError('Command did not exit with success.')

            self.stdout = stdout
            self.stderr = stderr


class babel(ExternalCommandTask):
    """
    Transpiles ES-next to ES-today.

    Available options are modeled after the command line options for the
    ``babel`` command line tool.
    """

    def process(self, inputs):
        cmd = ['babel']
        for t, c, m in inputs:
            yield t, [self.run_external(cmd, input=b''.join(c))], m


class coffee(ExternalCommandTask):
    """Compile Coffeescript to JavaScript."""

    def process(self, inputs):
        for t, c, m in inputs:
            cmd = ['coffee', '--print', '--stdio']
            output = self.run_external(cmd, input=b''.join(c))
            yield t.with_suffix('.js'), [output], m


class traceur(Task):
    """
    Transpiles ES6 JavaScript to ES5.

    Available options are modeled after the command line options for the
    ``traceur`` command line tool.

    Traceur supports two modes: compile as scripts or compile as modules.
    Both patterns are supported.
    """

    def __init__(self, *files,
                 block_binding=None,
                 modules=None,
                 out=None):
        self.block_binding = block_binding
        self.files = files
        self.modules = modules
        self.out = out

    def run_traceur(self):
        # Run traceur
        cmd = ['traceur']
        if self.modules:
            cmd += ['--modules={}'.format(self.modules)]
        if self.block_binding:
            cmd += ['--block-binding=true']
        if self.out:
            cmd += ['--out', self.out]
        if self.files:
            cmd += self.files

        command = ExternalCommand(cmd)
        command.require_empty_stdout = True
        command.require_empty_stderr = True
        command.require_success = True
        command.run()

    def process(self, inputs):
        tmp_dir, paths = write_to_tmp_dir(inputs)
        with chdir(tmp_dir):
            self.run_traceur()
        for t, c, m in walk(tmp_dir):
            if t not in paths:  # filter out any input files, returning only the new files
                yield t, c, m


class uglify(ExternalCommandTask):
    """Minify and obfuscate JavaScript files."""
    def __init__(self, *, compress=True, mangle=True, preamble=None, enclose=False):
        self.compress = compress
        self.enclose = enclose
        self.mangle = mangle
        self.preamble = preamble

    def process(self, inputs):
        for t, c, m in inputs:
            cmd = ['uglifyjs']
            if self.compress:
                cmd += ['--compress']
            if self.mangle:
                cmd += ['--mangle']
            if self.enclose:
                cmd += ['--enclose']
            if self.preamble:
                cmd += ['--preamble', self.preamble]
            output = self.run_external(cmd, input=b''.join(c))
            yield t, [output], m
